@@ -1,7 +1,8 @@
 #![forbid(unsafe_code)]
 #![doc(
     html_root_url = "https://docs.rs/pyo3-log/0.2.1/pyo3-log/",
-    test(attr(deny(warnings)))
+    test(attr(deny(warnings))),
+    test(attr(allow(unknown_lints, non_local_definitions)))
 )]
 #![warn(missing_docs)]
 
@@ -37,7 +38,7 @@
 //! }
 //!
 //! #[pymodule]
-//! fn my_module(_py: Python<'_>, m: &PyModule) -> PyResult<()> {
+//! fn my_module(m: Bound<'_, PyModule>) -> PyResult<()> {
 //!     pyo3_log::init();
 //!
 //!     m.add_wrapped(wrap_pyfunction!(log_something))?;
@@ -216,28 +217,45 @@ impl Default for Caching {
     }
 }
 
-#[derive(Clone, Debug)]
+#[derive(Debug)]
 struct CacheEntry {
     filter: LevelFilter,
     logger: PyObject,
 }
 
-#[derive(Clone, Debug, Default)]
+impl CacheEntry {
+    fn clone_ref(&self, py: Python<'_>) -> Self {
+        CacheEntry {
+            filter: self.filter,
+            logger: self.logger.clone_ref(py),
+        }
+    }
+}
+
+#[derive(Debug, Default)]
 struct CacheNode {
     local: Option<CacheEntry>,
     children: HashMap<String, Arc<CacheNode>>,
 }
 
 impl CacheNode {
-    fn store_to_cache_recursive<'a, P>(&self, mut path: P, entry: CacheEntry) -> Arc<Self>
+    fn store_to_cache_recursive<'a, P>(
+        &self,
+        py: Python<'_>,
+        mut path: P,
+        entry: CacheEntry,
+    ) -> Arc<Self>
     where
         P: Iterator<Item = &'a str>,
     {
-        let mut me = self.clone();
+        let mut me = CacheNode {
+            children: self.children.clone(),
+            local: self.local.as_ref().map(|e| e.clone_ref(py)),
+        };
         match path.next() {
             Some(segment) => {
                 let child = me.children.entry(segment.to_owned()).or_default();
-                *child = child.store_to_cache_recursive(path, entry);
+                *child = child.store_to_cache_recursive(py, path, entry);
             }
             None => me.local = Some(entry),
         }
@@ -252,7 +270,7 @@ impl CacheNode {
 ///
 /// It can be either created directly and then installed, passed to other aggregating log systems,
 /// or the [`init`] or [`try_init`] functions may be used if defaults are good enough.
-#[derive(Clone, Debug)]
+#[derive(Debug)]
 pub struct Logger {
     /// Filter used as a fallback if none of the `filters` match.
     top_filter: LevelFilter,
@@ -285,7 +303,7 @@ impl Logger {
     ///
     /// It defaults to having a filter for [`Debug`][LevelFilter::Debug].
     pub fn new(py: Python<'_>, caching: Caching) -> PyResult<Self> {
-        let logging = py.import("logging")?;
+        let logging = py.import_bound("logging")?;
         Ok(Self {
             top_filter: LevelFilter::Debug,
             filters: HashMap::new(),
@@ -399,10 +417,10 @@ impl Logger {
             .and_then(|node| node.local.as_ref())
             .map(|local| &local.logger);
         let (logger, cached) = match cached_logger {
-            Some(cached) => (cached.as_ref(py), true),
+            Some(cached) => (cached.bind(py).clone(), true),
             None => (
                 self.logging
-                    .as_ref(py)
+                    .bind(py)
                     .getattr("getLogger")?
                     .call1((&target,))?,
                 false,
@@ -410,7 +428,7 @@ impl Logger {
         };
         // We need to check for this ourselves. For some reason, the logger.handle does not check
         // it. And besides, we can save ourselves few python calls if it's turned off.
-        if is_enabled_for(logger, record.level())? {
+        if is_enabled_for(&logger, record.level())? {
             let none = py.None();
             // TODO: kv pairs, if enabled as a feature?
             let record = logger.call_method1(
@@ -421,8 +439,8 @@ impl Logger {
                     record.file(),
                     record.line().unwrap_or_default(),
                     msg,
-                    PyTuple::empty(py), // args
-                    &none,              // exc_info
+                    PyTuple::empty_bound(py), // args
+                    &none,                    // exc_info
                 ),
             )?;
             logger.call_method1("handle", (record,))?;
@@ -463,12 +481,12 @@ impl Logger {
         metadata.level() <= cache_filter && metadata.level() <= self.filter_for(metadata.target())
     }
 
-    fn store_to_cache(&self, target: &str, entry: CacheEntry) {
+    fn store_to_cache(&self, py: Python<'_>, target: &str, entry: CacheEntry) {
         let path = target.split("::");
 
         let orig = self.cache.load();
         // Construct a new cache structure and insert the new root.
-        let new = orig.store_to_cache_recursive(path, entry);
+        let new = orig.store_to_cache_recursive(py, path, entry);
         // Note: In case of collision, the cache update is lost. This is fine, as we simply lose a
         // tiny bit of performance and will cache the thing next time.
         //
@@ -496,32 +514,26 @@ impl Log for Logger {
     fn log(&self, record: &Record) {
         let cache = self.lookup(record.target());
 
-        let mut store_to_cache = None;
         if self.enabled_inner(record.metadata(), &cache) {
             Python::with_gil(|py| match self.log_inner(py, record, &cache) {
                 Ok(Some(logger)) => {
                     let filter = match self.caching {
                         Caching::Nothing => unreachable!(),
                         Caching::Loggers => LevelFilter::max(),
-                        Caching::LoggersAndLevels => {
-                            extract_max_level(py, &logger).unwrap_or_else(|e| {
+                        Caching::LoggersAndLevels => extract_max_level(logger.bind(py))
+                            .unwrap_or_else(|e| {
                                 e.print(py);
                                 LevelFilter::max()
-                            })
-                        }
+                            }),
                     };
-                    store_to_cache = Some((logger, filter));
+
+                    let entry = CacheEntry { filter, logger };
+                    self.store_to_cache(py, record.target(), entry);
                 }
                 Ok(None) => (),
                 Err(e) => e.print(py),
             })
         }
-        // Note: no more GIL here. Not needed for storing to cache.
-
-        if let Some((logger, filter)) = store_to_cache {
-            let entry = CacheEntry { filter, logger };
-            self.store_to_cache(record.target(), entry);
-        }
     }
 
     fn flush(&self) {}
@@ -537,14 +549,13 @@ fn map_level(level: Level) -> usize {
     }
 }
 
-fn is_enabled_for(logger: &PyAny, level: Level) -> PyResult<bool> {
+fn is_enabled_for(logger: &Bound<'_, PyAny>, level: Level) -> PyResult<bool> {
     let level = map_level(level);
-    logger.call_method1("isEnabledFor", (level,))?.is_true()
+    logger.call_method1("isEnabledFor", (level,))?.is_truthy()
 }
 
-fn extract_max_level(py: Python<'_>, logger: &PyObject) -> PyResult<LevelFilter> {
+fn extract_max_level(logger: &Bound<'_, PyAny>) -> PyResult<LevelFilter> {
     use Level::*;
-    let logger = logger.as_ref(py);
     for l in &[Trace, Debug, Info, Warn, Error] {
         if is_enabled_for(logger, *l)? {
             return Ok(l.to_level_filter());
